#ifndef DIETERICH_STATE_UPDATER_HH
#define DIETERICH_STATE_UPDATER_HH

#include "stateupdater.hh"

template <class ScalarVector, class Vector>
class DieterichStateUpdater : public StateUpdater<ScalarVector, Vector> {
public:
  DieterichStateUpdater(ScalarVector _logState_initial,
                        Dune::BitSetVector<1> const &_nodes, double _L);

  void virtual nextTimeStep();
  void virtual setup(double _tau);
  void virtual solve(Vector const &velocity_field);
  void virtual extractLogState(ScalarVector &);

private:
  ScalarVector logState_o;
  ScalarVector logState;
  Dune::BitSetVector<1> const &nodes;
  double const L;
  double tau;
};

template <class ScalarVector, class Vector>
DieterichStateUpdater<ScalarVector, Vector>::DieterichStateUpdater(
    ScalarVector _logState_initial, Dune::BitSetVector<1> const &_nodes,
    double _L)
    : logState(_logState_initial), nodes(_nodes), L(_L) {}

template <class ScalarVector, class Vector>
void DieterichStateUpdater<ScalarVector, Vector>::nextTimeStep() {
  logState_o = logState;
}

template <class ScalarVector, class Vector>
void DieterichStateUpdater<ScalarVector, Vector>::setup(double _tau) {
  tau = _tau;
}

/*
  Approximate (1-exp(-tV))/V, i.e. t*(exp(x) - 1)/x
  for x = -t*v close to 0

  with t, V >= 0, so that x should be non-positive
*/
double approximateMyPowerSeries(double V, double t) {
  double mVt = -V * t;
  if (mVt >= 0)
    return t;

  return -std::expm1(mVt) / V;
}

template <class ScalarVector, class Vector>
void DieterichStateUpdater<ScalarVector, Vector>::solve(
    Vector const &velocity_field) {
  for (size_t i = 0; i < nodes.size(); ++i) {
    if (not nodes[i][0])
      continue;

    double const VoL = velocity_field[i].two_norm() / L;
    // log( (1-exp(-tV))/V + exp(alpha_old - tV) ) with a
    // liftable singularity at V=0
    logState[i] = std::log(approximateMyPowerSeries(VoL, tau) +
                           std::exp(logState_o[i] - tau * VoL));
  }
}

template <class ScalarVector, class Vector>
void DieterichStateUpdater<ScalarVector, Vector>::extractLogState(
    ScalarVector &_logState) {
  _logState = logState;
}

#endif
